import datetime as dt
import logging

import hydropandas as hpd
import numpy as np
import pandas as pd

from .. import cache, util
from ..dims.layers import get_first_active_layer
from ..dims.resample import get_affine_mod_to_world

logger = logging.getLogger(__name__)


@cache.cache_netcdf
def get_recharge(ds, method="linear"):
    """add multiple recharge packages to the groundwater flow model with knmi
    data by following these steps:

       1. check for each cell (structured or vertex) which knmi measurement
          stations (prec and evap) are the closest.
       2. download precipitation and evaporation data for all knmi stations that
          were found at 1
       3. create a recharge package in which each cell has a reference to a
          timeseries. Timeseries are created for each unique combination of
          precipitation and evaporation. The following packages are created:
            a. the rch package itself in which cells with the same
               precipitation and evaporation stations are defined. This
               package also refers to all the time series package (see b).
            b. the time series packages in which the recharge flux is defined
               for the time steps of the model. Each package contains the
               time series for one or more cels (defined in a).

    Supports structured and unstructred datasets.

    Parameters
    ----------
    ds : xr.DataSet
        dataset containing relevant model grid information
    method : bool, optional
        If 'linear', calculate recharge by subtracting evaporation from precipitation.
        If 'separate', add precipitation as 'recharge' and evaporation as 'evaporation'.
        The defaults is 'linear'.

    Returns
    -------
    ds : xr.DataSet
        dataset with spatial model data including the rch raster
    """

    start = pd.Timestamp(ds.time.attrs["start"])
    # include the end day in the time series.
    end = pd.Timestamp(ds.time.data[-1]) + pd.Timedelta(1, "D")

    ds_out = util.get_ds_empty(ds)
    ds_out.attrs["gridtype"] = ds.gridtype

    # get recharge data array
    if ds.gridtype == "structured":
        dims = ("y", "x")
    elif ds.gridtype == "vertex":
        dims = ("icell2d",)
    if not ds.time.steady_state:
        dims = dims + ("time",)

    shape = [len(ds_out[dim]) for dim in dims]

    locations, oc_knmi_prec, oc_knmi_evap = get_knmi_at_locations(
        ds, start=start, end=end
    )

    # add closest precipitation and evaporation measurement station to each cell
    locations[["prec_point", "distance"]] = locations.geo.get_nearest_point(
        oc_knmi_prec
    )
    locations[["evap_point", "distance"]] = locations.geo.get_nearest_point(
        oc_knmi_evap
    )

    if method in ["linear"]:
        ds_out["recharge"] = dims, np.zeros(shape)

        # find unique combination of precipitation and evaporation station
        unique_combinations = locations.drop_duplicates(["prec_point", "evap_point"])[
            ["prec_point", "evap_point"]
        ].values

        for prec_stn, evap_stn in unique_combinations:
            # get locations with the same prec and evap station
            loc_sel = locations.loc[
                (locations["prec_point"] == prec_stn)
                & (locations["evap_point"] == evap_stn)
            ]

            # calculate recharge time series
            prec = oc_knmi_prec.loc[prec_stn, "obs"]["RD"]
            evap = oc_knmi_evap.loc[evap_stn, "obs"]["EV24"]
            ts = (prec - evap).dropna()
            ts.name = f"{prec.name}-{evap.name}"

            _add_ts_to_ds(ts, loc_sel, "recharge", ds_out)

    elif method == "separate":
        ds_out["recharge"] = dims, np.zeros(shape)
        for stn in locations["prec_point"].unique():
            ts = oc_knmi_prec.loc[stn, "obs"]["RD"]
            loc_sel = locations.loc[(locations["prec_point"] == stn)]
            _add_ts_to_ds(ts, loc_sel, "recharge", ds_out)
        ds_out["evaporation"] = dims, np.zeros(shape)
        for stn in locations["evap_point"].unique():
            ts = oc_knmi_evap.loc[stn, "obs"]["EV24"]
            loc_sel = locations.loc[(locations["evap_point"] == stn)]
            _add_ts_to_ds(ts, loc_sel, "evaporation", ds_out)
    else:
        raise (Exception(f"Unknown method: {method}"))
    for datavar in ds_out:
        ds_out[datavar].attrs["source"] = "KNMI"
        ds_out[datavar].attrs["date"] = dt.datetime.now().strftime("%Y%m%d")
        ds_out[datavar].attrs["units"] = "m/day"

    return ds_out


def _add_ts_to_ds(timeseries, loc_sel, variable, ds):
    """Add a timeseries to a variable at location loc_sel in model DataSet."""
    end = pd.Timestamp(ds.time.data[-1]) + pd.Timedelta(1, "D")
    if timeseries.index[-1] < pd.Timestamp(ds.time.data[-1]):
        raise ValueError(f"no recharge available at {timeseries.name} for date {end}")

    # fill recharge data array
    if ds.time.steady_state:
        rch_average = timeseries.mean()
        if ds.gridtype == "structured":
            # add data to ds
            for row, col in zip(loc_sel.row, loc_sel.col):
                ds[variable].data[row, col] = rch_average
        elif ds.gridtype == "vertex":
            # add data to ds
            ds[variable].loc[loc_sel.index] = rch_average
    else:
        model_recharge = pd.Series(index=ds.time.data, dtype=float)
        for j, ts in enumerate(model_recharge.index):
            if j < (len(model_recharge) - 1):
                model_recharge.loc[ts] = (
                    timeseries.loc[ts : model_recharge.index[j + 1]].iloc[:-1].mean()
                )
            else:

                model_recharge.loc[ts] = timeseries.loc[ts:end].iloc[:-1].mean()

        # add data to ds
        if ds.gridtype == "structured":
            for row, col in zip(loc_sel.row, loc_sel.col):
                ds[variable].data[row, col, :] = model_recharge.values

        elif ds.gridtype == "vertex":
            ds[variable].loc[loc_sel.index, :] = model_recharge.values


def get_locations_vertex(ds):
    """get dataframe with the locations of the grid cells of a vertex grid.

    Parameters
    ----------
    ds : xr.DataSet
        dataset containing relevant model grid information

    Returns
    -------
    locations : pandas DataFrame
        DataFrame with the locations of all active grid cells.
        includes the columns: x, y and layer
    """
    # get active locations
    fal = get_first_active_layer(ds)
    icell2d_active = np.where(fal != fal.attrs["_FillValue"])[0]

    # create dataframe from active locations
    x = ds["x"].sel(icell2d=icell2d_active)
    y = ds["y"].sel(icell2d=icell2d_active)
    if "angrot" in ds.attrs and ds.attrs["angrot"] != 0.0:
        # transform coordinates into real-world coordinates
        affine = get_affine_mod_to_world(ds)
        x, y = affine * (x, y)
    layer = fal.sel(icell2d=icell2d_active)
    locations = pd.DataFrame(
        index=icell2d_active, data={"x": x, "y": y, "layer": layer}
    )
    locations = hpd.ObsCollection(locations)

    return locations


def get_locations_structured(ds):
    """get dataframe with the locations of the grid cells of a structured grid.

    Parameters
    ----------
    ds : xr.DataSet
        dataset containing relevant model grid information

    Returns
    -------
    locations : pandas DataFrame
        DataFrame with the locations of all active grid cells.
        includes the columns: x, y, row, col and layer
    """

    # store x and y mids in locations of active cells
    fal = get_first_active_layer(ds)
    rows, columns = np.where(fal != fal.attrs["_FillValue"])
    x = np.array([ds["x"].data[col] for col in columns])
    y = np.array([ds["y"].data[row] for row in rows])
    if "angrot" in ds.attrs and ds.attrs["angrot"] != 0.0:
        # transform coordinates into real-world coordinates
        affine = get_affine_mod_to_world(ds)
        x, y = affine * (x, y)
    layers = [fal.data[row, col] for row, col in zip(rows, columns)]

    locations = hpd.ObsCollection(
        pd.DataFrame(
            data={"x": x, "y": y, "row": rows, "col": columns, "layer": layers}
        )
    )

    return locations


def get_knmi_at_locations(ds, start="2010", end=None):
    """get knmi data at the locations of the active grid cells in ds.

    Parameters
    ----------
    ds : xr.DataSet
        dataset containing relevant model grid information
    start : str or datetime, optional
        start date of measurements that you want, The default is '2010'.
    end :  str or datetime, optional
        end date of measurements that you want, The default is None.

    Raises
    ------
    ValueError
        wrong grid type specified.

    Returns
    -------
    locations : pandas DataFrame
        DataFrame with the locations of all active grid cells.
    oc_knmi_prec : hydropandas.ObsCollection
        ObsCollection with knmi data of the precipitation stations.
    oc_knmi_evap : hydropandas.ObsCollection
        ObsCollection with knmi data of the evaporation stations.
    """
    # get locations
    if ds.gridtype == "structured":
        locations = get_locations_structured(ds)
    elif ds.gridtype == "vertex":
        locations = get_locations_vertex(ds)
    else:
        raise ValueError("gridtype should be structured or vertex")

    # get knmi data stations closest to any grid cell
    oc_knmi_prec = hpd.ObsCollection.from_knmi(
        locations=locations, starts=[start], ends=[end], meteo_vars=["RD"]
    )

    oc_knmi_evap = hpd.ObsCollection.from_knmi(
        locations=locations, starts=[start], ends=[end], meteo_vars=["EV24"]
    )

    return locations, oc_knmi_prec, oc_knmi_evap
